#include <ProtocolManager.h>

#include <boost/date_time.hpp>
#include <boost/filesystem.hpp>

namespace DataExporter_ns
{

//==============================================================================
//      ProtocolManager::ProtocolManager()
//==============================================================================
ProtocolManager::ProtocolManager(Tango::DeviceImpl* deviceImpl_p,
    Configuration::SP configuration_sp, DBManager::SP dBManager_sp) :
    Tango::LogAdapter(deviceImpl_p), m_configuration_sp(configuration_sp),
    m_dBManager_sp(dBManager_sp)
{
    DEBUG_STREAM << "ProtocolManager::ProtocolManager()" << endl;

    m_isAuthorised = false;
    m_isValidated = false;
    m_isTransferRequest = false;
}

//==============================================================================
//      ProtocolManager::~ProtocolManager()
//==============================================================================
ProtocolManager::~ProtocolManager()
{
    DEBUG_STREAM << "ProtocolManager::~ProtocolManager()" << endl;
}

//==============================================================================
//      ProtocolManager::create()
//==============================================================================
ProtocolManager::SP ProtocolManager::create(Tango::DeviceImpl* deviceImpl_p,
    Configuration::SP configuration_sp, DBManager::SP dBManager_sp)
{
    ProtocolManager::SP d_sp(new ProtocolManager(deviceImpl_p, configuration_sp,
        dBManager_sp), ProtocolManager::Deleter());

    return d_sp;
}

//==============================================================================
//      ProtocolManager::setRemoteEndpoint()
//==============================================================================
void ProtocolManager::setRemoteEndpoint(std::string remoteEndpoint)
{
    DEBUG_STREAM << "ProtocolManager::setRemoteEndpoint()" << endl;

    m_remoteEndpoint = remoteEndpoint;
}

//==============================================================================
//      ProtocolManager::setRemoteEndpoint()
//==============================================================================
ResponseSP ProtocolManager::prepareResponse(RequestSP request_sp)
        throw(std::runtime_error)
{
    DEBUG_STREAM << "ProtocolManager::prepareResponse()" << endl;

    if(!request_sp->IsInitialized())
        throw std::runtime_error("Not initialized request!");

    ResponseSP response_sp;

    switch(request_sp->type())
    {
        case Request::AUTHORIZATION:
        {
            response_sp = prepareAuthroisation(request_sp);
            break;
        }
        case Request::VALIDATION:
        {
            response_sp = prepareValidation(request_sp);
            break;
        }
        case Request::TRANSFER:
        {
            response_sp = prepareTransfer(request_sp);
            break;
        }
        case Request::KEEPALIVE:
        {
            response_sp = prepareKeepAlive(request_sp);
            break;
        }
        default:
            throw std::runtime_error("Unknown request type!");
    }

    if(!response_sp->IsInitialized())
        throw std::runtime_error("Not initialized response!");

    return response_sp;
}

//==============================================================================
//      ProtocolManager::isTransferRequest()
//==============================================================================
bool ProtocolManager::isTransferRequest()
{
    DEBUG_STREAM << "ProtocolManager::isTransferRequest()" << endl;

    return m_isTransferRequest;
}

//==============================================================================
//      ProtocolManager::getFilePath()
//==============================================================================
std::string ProtocolManager::getFilePath()
{
    DEBUG_STREAM << "ProtocolManager::getFilePath()" << endl;

    return m_filePath;
}

//==============================================================================
//      ProtocolManager::getFileSize()
//==============================================================================
int ProtocolManager::getFileSize()
{
    DEBUG_STREAM << "ProtocolManager::getFileSize()" << endl;

    return m_fileSize;
}

//==============================================================================
//      ProtocolManager::prepareAuthroisation()
//==============================================================================
ResponseSP ProtocolManager::prepareAuthroisation(RequestSP request_sp)
{
    DEBUG_STREAM << "ProtocolManager::prepareAuthroisation()" << endl;

    ResponseSP response_sp(new Response());

    response_sp->set_type(Response::AUTHORIZATION);

    Response::Authorization* authResp = response_sp->mutable_authorization();

    if(!m_isAuthorised)
    {
        const Request::Authorization& authReq = request_sp->authorization();
        std::string username =  authReq.username();
        std::string password = authReq.password();

        if(m_configuration_sp->isUserAuthorized(username, password))
        {
            INFO_STREAM << "ProtocolManager::prepareAuthroisation() "
                << "Authorization accepted from " << m_remoteEndpoint << endl;

            m_isAuthorised = true;

            authResp->set_state(Response::Authorization::ACCEPTED);
            authResp->set_status("Authorization accepted");
        }
        else
        {
            WARN_STREAM << "ProtocolManager::prepareAuthroisation() "
                << "Invalid username or password from " << m_remoteEndpoint << endl;

            m_isAuthorised = false;

            authResp->set_state(Response::Authorization::REJECTED);
            authResp->set_status("Invalid username or password");
        }
    }
    else
    {
        WARN_STREAM << "ProtocolManager::prepareAuthroisation() "
            << "Already authorized from " << m_remoteEndpoint << endl;

        authResp->set_state(Response::Authorization::REJECTED);
        authResp->set_status("Already authorized");
    }

    return response_sp;
}

//==============================================================================
//      ProtocolManager::prepareValidation()
//==============================================================================
ResponseSP ProtocolManager::prepareValidation(RequestSP request_sp)
{
    DEBUG_STREAM << "ProtocolManager::prepareValidation()" << endl;

    ResponseSP response_sp(new Response());

    response_sp->set_type(Response::VALIDATION);

    Response::Validation* validationRes = response_sp->mutable_validation();

    if(m_isAuthorised)
    {
        if(!m_isValidated)
        {
            const Request::Validation& validationReq = request_sp->validation();

            m_validatedSchema = validationReq.schema();
            m_validatedTable =  validationReq.table();

            if(m_configuration_sp->isTableExported(m_validatedSchema, m_validatedTable))
            {
                INFO_STREAM << "ProtocolManager::prepareValidation() "
                    << "Validation accepted for " << m_validatedSchema << "."
                    << m_validatedTable << " from " << m_remoteEndpoint << endl;

                m_isValidated = true;

                validationRes->set_state(Response::Validation::ACCEPTED);
                validationRes->set_status("Table validated");
            }
            else
            {
                WARN_STREAM << "ProtocolManager::prepareValidation() "
                    << "Not exported table from " << m_remoteEndpoint << endl;

                validationRes->set_state(Response::Validation::REJECTED);
                validationRes->set_status("Not exported table");
            }
        }
        else
        {
            WARN_STREAM << "ProtocolManager::prepareValidation() "
                << "Already validated from " << m_remoteEndpoint << endl;

            validationRes->set_state(Response::Validation::REJECTED);
            validationRes->set_status("Already validated");
        }
    }
    else
    {
        WARN_STREAM << "ProtocolManager::prepareValidation() "
            << "Not authorised from " << m_remoteEndpoint << endl;

        validationRes->set_state(Response::Validation::REJECTED);
        validationRes->set_status("Not authorised");
    }

    return response_sp;
}

//==============================================================================
//      ProtocolManager::prepareTransfer()
//==============================================================================
ResponseSP ProtocolManager::prepareTransfer(RequestSP request_sp)
{
    DEBUG_STREAM << "ProtocolManager::prepareTransfer()" << endl;

    ResponseSP response_sp(new Response());

    response_sp->set_type(Response::TRANSFER);

    Response::Transfer* transferRes = response_sp->mutable_transfer();

    if(m_isAuthorised)
    {
        if(m_isValidated)
        {
            const Request::Transfer& transferReq = request_sp->transfer();

            int fileVersion = transferReq.file_version();
            std::string fileName = transferReq.file_name();

            try
            {
                m_isTransferRequest = false;

                DBManager::FileTuple fileTuple =
                    m_dBManager_sp->retrieveFileInfo(m_validatedSchema,
                    m_validatedTable, fileVersion, fileName);

                if(!fileTuple.get<0>() || !fileTuple.get<1>())
                {
                    std::stringstream errorStream;
                    errorStream << "File " << fileName << " version "
                        << fileVersion << " not yet downloaded in storage";
                    throw std::runtime_error(errorStream.str());
                }

                std::string storagePath = fileTuple.get<0>().get();
                boost::filesystem::path absPath(storagePath);

                std::string filePath = fileTuple.get<1>().get();
                boost::filesystem::path relPath(filePath);

                absPath /= relPath;

                std::stringstream pathStream;
                pathStream << "/" << fileVersion << "/" << fileName;
                boost::filesystem::path lastPath(pathStream.str());

                absPath /= lastPath;

                DEBUG_STREAM << "FILE: " << absPath.string() << endl;

                if(!boost::filesystem::is_regular_file(absPath))
                {
                    std::stringstream errorStream;
                    errorStream << "File " << fileName << " version "
                        << fileVersion << " not exists in storage";
                    throw std::runtime_error(errorStream.str());
                }

                transferRes->set_file_path(filePath);
                transferRes->set_file_version(fileVersion);
                transferRes->set_file_name(fileName);
                transferRes->set_size(boost::filesystem::file_size(absPath));

                transferRes->set_state(Response::Transfer::ACCEPTED);
                transferRes->set_status("File found");

                m_isTransferRequest = true;
                m_filePath = absPath.string();
                m_fileSize = boost::filesystem::file_size(absPath);
            }
            catch(std::exception& ex)
            {
                WARN_STREAM << "ProtocolManager::prepareMetadata() "
                    << ex.what() << " from " << m_remoteEndpoint << endl;

                transferRes->set_state(Response::Transfer::REJECTED);
                transferRes->set_status(ex.what());
            }
        }
        else
        {
            WARN_STREAM << "ProtocolManager::prepareMetadata() "
                << "Not validated from " << m_remoteEndpoint << endl;

            transferRes->set_state(Response::Transfer::REJECTED);
            transferRes->set_status("Not validated");
        }
    }
    else
    {
        WARN_STREAM << "ProtocolManager::prepareData() "
            << "Not authorised from " << m_remoteEndpoint << endl;

        transferRes->set_state(Response::Transfer::REJECTED);
        transferRes->set_status("Not authorised");
    }

    return response_sp;
}

//==============================================================================
//      ProtocolManager::prepareKeepAlive()
//==============================================================================
ResponseSP ProtocolManager::prepareKeepAlive(RequestSP request_sp)
{
    ResponseSP response_sp(new Response());

    response_sp->set_type(Response::KEEPALIVE);

    m_isTransferRequest = false;

    m_filePath.clear();

    m_fileSize = 0;

    return response_sp;
}

}   //namespace
